/***************************************************************************
                          stcpserver.cpp  -  really cool tcp server
                             -------------------
    begin                : 01-07-2003
    copyright            : (C) 2003 by Andr Stein
    email                : stone@steinsoft.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <iostream>
#include <unistd.h>
#include <string>
#include <fstream>
#include <signal.h>

using namespace std;

/**
  GLOBALS
 ***/

//4K receive buffer
#define BUFFER_LENGTH 4096

//server port to be used
unsigned int SERVER_PORT = 8888;
//file we stream to
char STREAM_FILE[ 256 ];
//file..
ofstream file;
//output status messages
bool silent = false;


/**
  FUNCTIONS
 ***/

void showHelpScreen()
{
	cout << "tcpserver v0.1" << endl
	<< "  programmed by Andr Stein <www.steinsoft.net>" << endl
	<< "  simple tcp daemon that outputs or redirects data it gets" << endl
	<< endl
	<< " Options:" << endl
	<< "   --help\t\tthis one" << endl
	<< "   --port n\t\tset listening port to n (default: 8888)" << endl
	<< "   --file file\t\tset destination file (default: stdout)" << endl
	<< "   --silent\t\tbe silent and don't print status messages" << endl;
}


void processCommandLine( int argc, char* argv[] )
{
	//not enough options so that it's worth
	//processing
	if ( argc <= 1 )
	{
		return ;
	}

	for ( int i = 1; i < argc; ++i )
	{
		if ( !strcmp( argv[ i ], "--help" ) )
		{
			showHelpScreen();
			exit( 0 );
		}
		else if ( !strcmp( argv[ i ], "--port" ) )
		{
			SERVER_PORT = atoi( argv[ i + 1 ] );
			continue;
		}
		else if ( !strcmp( argv[ i ], "--file" ) )
		{
			strncpy( STREAM_FILE, argv[ i + 1 ], 256 );
			continue;
		}
		else if ( !strcmp( argv[ i ], "--silent" ) )
		{
			silent = true;
		}
	}
}

//clean exit
void exitHandler( int )
{
	//close file to prevent any data lost
	file.close();

	cout << " exiting." << endl;
}

int main( int argc, char* argv[] )
{
	unsigned int serverSocket,
		      clientSocket,
		      length;
	int status;
	char buffer[BUFFER_LENGTH];

	strcpy( STREAM_FILE, "stdout" );

	//process commandline first..
	processCommandLine( argc, argv );
	//set signal handler..
	signal( SIGTERM, exitHandler );

	if ( !silent )
		cout << "tcpserver listening on port " << SERVER_PORT << endl
		<< "  output directed to " << STREAM_FILE << endl;


	/**
	  setup file
	 **/
	//because we're lazy, use stdout device file
	if ( !strcmp( STREAM_FILE, "stdout" ) )
		strcpy( STREAM_FILE, "/dev/stdout" );

	file.open( STREAM_FILE, ios::out | ios::app );

	/**
	  setup socket and thus init network connection
	 **/
	sockaddr_in local,
		     client;

	serverSocket = socket( AF_INET, SOCK_STREAM, 0 );

	local.sin_family = AF_INET;
	local.sin_port = htons( SERVER_PORT );
	local.sin_addr.s_addr = htonl( INADDR_ANY );


	if ( bind( serverSocket, ( sockaddr* ) & local, sizeof( local ) ) < 0 )
	{
		cerr << "error: couldn't bind socket." << endl;
		return 0;
	}

	//listen for clients
	listen( serverSocket, 5 );

	for ( ;; )
	{
		length = sizeof( client );
		clientSocket = accept( serverSocket, ( sockaddr* ) & client, ( socklen_t* ) & length );

		if ( clientSocket < 0 )
		{
			cerr << "error: could not accept connection from client." << endl;
		}

		//connected
		if ( !silent )
			cout << "  client connected -- " << inet_ntoa( client.sin_addr ) << endl
			<< "    beginning to stream" << endl;

		//GET'N STREAM
		for ( ;; )
		{
			status = recv( clientSocket, buffer, BUFFER_LENGTH, 0 );

			if ( status < 0 )
			{
				cerr << "error: cannot receive data from client." << endl << endl;
				break;
			}
			else if ( !status )
			{
				cerr << "error: connection closed by client." << endl << endl;
				close( clientSocket );
				break;
			}

			//write actual message using length
			//returned form recv(..)
			file.write(buffer, status);
			//flush file buffer to prevent data lost
			file.flush();
		}
	}

}


